/* Precomputations for numerical integration on [a, oo[
 * Behaviour of f at infinity:
 *  fast > 0: exponential decrease as exp(-fast*x)
 *  -2 < fast < -1: slow decrease, as x^fast
 *  fast <= -2, fast = 0, or omitted: as x^(-2) or better */
Inta_ooinit(fast = 0, mul = 1) =
{
  if (fast > 0, return (Inta_oofastinit(mul)));
  if (fast < 0 && fast >= -1, error("impossible fast"));
  if (fast <= -2, fast = 0);
  return (Inta_ooslowinit(fast, mul));
}

Inta_ooslowinit(fast = 0, mul = 1) =
{
  my(B, N, h, b, e1, q, Xp, Wp, Xm, Wm);
  b = if (fast, -2 / (fast + 1), 2); \\ > 0
  B = getlocalbitprec(); N = IntgetN(B, 3.49 / mul);
  localbitprec(B + 32); h = log(2*Pi*N/log(N)) / N;
  Xp = vector(N); Xm = vector(N);
  Wp = vector(N); Wm = vector(N); e1 = exp(h); q = 1;
  for (m = 1, N,
    my(t);
    q *= e1; t = (q + 1 / q)/2; \\ exp(mh), cosh(mh)
    Xp[m] = exp(b * (q - t));
    Xm[m] = 1 / Xp[m];
    Wp[m] = Xp[m] * t;
    Wm[m] = Xm[m] * t;
    if (exponent(Xm[m]) < -B * b / 2 - 64,
      Xp = Xp[1..m]; Wp = Wp[1..m];
      Xm = Xm[1..m]; Wm = Wm[1..m]; break));
  return (bitprecision([b * h, Xp, Wp, Xm, Wm], B + 32));
}

Inta_oofastinit(mul = 1) =
{ my(B = getlocalbitprec(), N, h, e1, vexph, Xp, Xm, Wp, Wm);

  N = IntgetN(B, 5.85 / mul);
  localbitprec(B + 32); h = log(2*Pi*N/log(N)) / N;
  e1 = exp(h); vexph = powers(e1, N-1, e1);
  Xp = vector(N); Wp = vector(N);
  Xm = vector(N); Wm = vector(N);
  for (m = 1, N,
    my(vm = vexph[m], vmi = 1 / vm);
    Xp[m] = exp(m * h - vmi);
    Xm[m] = exp(-(m * h + vm));
    Wp[m] = Xp[m] * (1 + vmi);
    Wm[m] = Xm[m] * (1 + vm);
    if (exponent(Xm[m]) < -B - 64,
      my(b = exponent(Wp[m]) + B);
      if (Xp[m] > 0.694 * b + 1,
        Xp = Xp[1..m]; Wp = Wp[1..m];
        Xm = Xm[1..m]; Wm = Wm[1..m]); break));
  return (bitprecision([h, Xp, Wp, Xm, Wm], B + 32));
}

/* auxiliary sum */
Inta_oosum(f, a, Xp, Wp, Xm, Wm) =
  sum(m = 1, #Xp, Wp[m] * f(a+Xp[m]) + Wm[m] * f(a+Xm[m]));

/* Numerical integration of a closure f on [a, oo[, with
 * precomputed XW from Inta_ooinit, otherwise recomputed. */
Inta_oo(f, a = 1, fast = 0, XW = 0) =
{
  if (fast > 0, return (Inta_oofast(f, a, fast, XW)));
  if (!XW, XW = Inta_ooinit(fast));
  my(S, [bh, Xp, Wp, Xm, Wm] = XW);
  if (a <= 1, S = f(a + 1);
            , S = f(2*a); Xp *= a; Xm *= a; bh *= a);
  return (bh * (S + Inta_oosum(f, a, Xp, Wp, Xm, Wm)));
}

/* Assume f goes to 0 like exp(-fast * x). */
Inta_oofast(f, a = 1, fast = 1, XW = 0) =
{
  if (!XW, XW = Inta_oofastinit());
  my(S, [h, Xp, Wp, Xm, Wm] = XW, e1 = exp(-1) / fast);
  if (fast != 1, Xp /= fast; Xm /= fast);
  if (a <= 1, S = f(a + e1);
            , S = f(a * (1 + e1)); Xp *= a; Xm *= a; h *= a);
  S = 2 * e1 * S + Inta_oosum(f, a, Xp, Wp, Xm, Wm) / fast;
  return (h * S);
}

/* Assume f goes to 0 like exp(-fast * x^{1/2}). */
Inta_oohalffast(f, a = 1, fast = 1, XW = 0) =
{
  if (!XW, XW = Inta_oofastinit());
  2 * Inta_oofast(x -> x * f(x^2), sqrt(a), fast, XW);
}
