/* Generic Programs. */

/* Driver programs for numerical integration. Reduce to an
 * interval [1, N]. Here, [a, b] can be any compact complex
 * interval. Poles can be given either as complex numbers,
 * as polar parts, or as a minimal distance r. */

/* Sample initialization: Gauss-Legendre (d = 4), which reduces
 * to [0, 1] for integration on compact intervals. */
Intgeninit() =
{ my([X, W] = IntGaussLegendreinit(), n = 2 * #X, d = 4);
  /* Note: X contains only the positive nodes in ]-1, 1[ */

  X = vector(n, i, 1 + (-1)^i * X[(i + 1) \ 2]) / 2;
  W = vector(n, i, W[(i + 1) \ 2]) / 2;
  return ([X, W, d]);
}

/* Basic formula. */
Intgen0(f, a, b, XW) =
  my(c = b - a); c * SumXW(x -> f(a + c * x), XW);

/* Sum basic formula over a partition A of an interval. */
Intgensum(f, A, XW) =
  sum(j = 1, #A-1, Intgen0(f, A[j], A[j+1], XW));

/* Poles of have distance at least r to [1, N].
 * Subdivide [1, N] in equal pieces */
Intgenpolesmin(f, r, N, XW) =
{ my(B = getlocalbitprec(), n = #XW[1], d = XW[3], k, h);

  k = ceil((N - 1) * 2^(B/n) / (r * d));
  h = (N - 1) / k; /* equal pieces */
  return (Intgensum(f, vector(k+1, j, 1+(j-1)*h), XW));
}

/* Poles of f satisfy real(z) <= 1 and |z-1| >= 1.
 * Subdivide [1, N] using a geometric progression */
Intgenpoles1(f, N, XW) =
{ my(B = getlocalbitprec(), n = #XW[1], d = XW[3], k);

  k = ceil(log(N) / log(d * 2^(-B/n) + 1));
  return (Intgensum(f, powers(N^(1/k), k), XW));
}

/* Approximate (n!)^(1/n) */
Intgennsure(XW) = my(n = #XW[1]); return (n * exp(-1));

/* Case where poles can be ignored. */
Intgennopoles(f,N,XW) =
  Intgenpolesmin(f, Intgennsure(XW), N, XW);

/* Case where the polar part is known. */
Intgenpolar(f, F, N, XW) =
{ my(g = t -> f(t) - subst(F, variable(F), t));
  return (Intgennopoles(g, N, XW) + Intrat(F, 1, N));
}

/* Main driver routine on [a, b]. */
Intgen(f, a, b, XW = 0, poles = []) =
{ my(eith, c = b - a, N = abs(c) + 1);

  if (c == 0, return (0.));
  if (!XW, XW = Intgeninit());
  eith = if (imag(c), exp(I * arg(c))
                    , sign(real(c)));
  (f = x -> f(eith * (x-1) + a)); /* now on [1, N] */
  if (type(poles) != "t_VEC", /* minimal distance r > 0 */
    return (eith * Intgenpolesmin(f, poles, N, XW)));
  if (#poles == 0,
    return (eith * Intgennopoles(f, N, XW)));
  my(polar = 1);
  poles = vector(#poles, j,
      my(p = poles[j]);
      if (type(p) == "t_RFRAC",
        my(vx = variable(p));
        subst(p, vx, eith * (vx - 1) + a)
      , /* else */
        polar = 0;
        1 + conj(eith) * (p - a)));
  if (polar, /* all polar parts are known */
    return (eith * Intgenpolar(f, vecsum(poles), N, XW)));
  /* convert poles to list of complex numbers */
  poles = concat(vector(#poles, j,
    my(a = poles[j]);
    if (type(a) == "t_RFRAC", polroots(denominator(a)), [a])));
  return (eith * Intgenpoles(f, N, XW, poles));
}

/* Distance of complex z to [1, N]. */
Distpole(N,z) =
{ my(x = real(z));
  if (x < 1, return (abs(z - 1)));
  if (x > N, return (abs(z - N)));
  return (abs(imag(z)));
}

/* General case, on [1, N]; poles are complex numbers */
Intgenpoles(f, N, XW, poles) =
{ my(S, r, p1, pN, nsure = Intgennsure(XW));

  if (vecmin([Distpole(N,z) | z <- poles]) >= nsure,
    return (Intgenpolesmin(f, nsure, N, XW)));
  p1 = pN = List([nsure]);
  foreach (poles, a,
    my(x = real(a));
    if (x <= 1 && abs(a - 1) >= 0.1, listput(~p1, abs(a - 1)),
        x >= N && abs(a - N) >= 0.1, listput(~pN, abs(a - N)),
        error("please give polar parts in Intgen")));
  if (#pN == 1, r = vecmin(Vec(p1)),
      #p1 == 1, r = vecmin(Vec(pN)); f = x -> f(N+1-x),
      error("please give polar parts in Intgen"));
  /* Poles of f satisfy real(z) <= 1 and |z-1| >= r >= 0.1 */
  if (r >= 1, return (Intgenpoles1(f, N, XW)));
  /* A pole is close to 1 */
  S = r * Intgenpoles1(x -> f(r * (x - 1) + 1),
                       1 + min(N - 1, 1) / r, XW);
  if (N > 2, S += Intgenpoles1(x -> f(x + 1), N - 1, XW));
  return (S);
}
