/* 3-term recursion coefficients a, b attached to orthogonal
 * polynomial P_N are given as vectors
 *   a = [a(0), ..., a(N-1)], b = [b(1), ..., b(N-1)]
 * (they may contain further coefficients) */

/* Let r be a (complex or formal) value and a, b be the 3-term
 * recursion coefficients vectors. Let f be the attached MONIC
 * orthogonal polynomial f = P_N / pollead(P_N).
 * If k > 0, return
 *   [f(r), f'(r), f''(r) / 2, ..., f^(k-1)(r) / (k-1)!]
 * If k <= 0, only return f^(-k)(r) / (-k)! */
Orthopolyeval(a, b, N, r, k = 0) =
{ my(K = if (k > 0, k, 1 - k), u0 = vector(K), u1 = vector(K));

  u0[1] = 1; u1[1] = r - a[1]; if (K > 1, u1[2] = 1);
  for (m = 1, N - 1,
    my(u2 = (r - a[m+1]) * u1 - b[m] * u0);
    u0 = u1; u1 = u2 + concat(0, u1[1..K-1]));
  return (if (k <= 0, u1[K], u1));
}

/* Adaptation of Newtonrefine. */
OrthoNewtonrefine(a, b, N, r, B) =
{ my(c, p, p1, p2, n, e);

  [p, p1, p2] = Orthopolyeval(a, b, N, r, 3);
  e = max(exponent(p2 / p1) - 1, 0);
  if (e >= 63, error("Insuficcient accuracy in Newtonrefine"));
  n = 1 + exponent((B + 32 - e) \ (64 - e));
  c = 1 + e + (B - e) >> n;
  for (j = 1, n,
    c = 2*c - e; r = bitprecision(r, c);
    [p, p1] = Orthopolyeval(a, b, N, r, 2);
    r -= p / p1);
  return (bitprecision(r, B));
}

/* Given the 3-term recursion coefficients a, b as closures
 * or the corresponding vectors with N and N-1 components,
 * compute the roots of the orthogonal polynomial P_N.
 * If a = 0, only half the roots are computed */
Orthopolyroots(a, b, N) =
{ my(B = getlocalbitprec(), r, R, sym, vsb);

  if (type(a) == "t_CLOSURE", a = vector(N, n, a(n)));
  if (type(b) == "t_CLOSURE", b = vector(N - 1, n, b(n)));
  localbitprec(96); vsb = sqrt(b); localbitprec(64);
  r = Tqli(a, vsb); /* roots with 64 bit accuracy */
  sym = (#r < N); /* complete by symmetry ? */
  R = vector(N);
  for (i = 1, #r,
    R[i] = OrthoNewtonrefine(a, b, N, r[i], B + 32);
    if (sym, R[N + 1 - i] = -R[i]));
  return (R);
}
