/* Squarefree factorization [[P1,c1],...,[Pk,ck]] of exact
 * polynomial P = P1^c1 ... Pk^ck; characteristic is 0 */
PolSQF(P) =
{ my(SQF = List(), T, V, c);

  T = gcd(P, P'); V = P \ T; c = 0;
  while (poldegree(V),
    my(W = gcd(T, V), P = V \ W);
    c++; if (poldegree(P), listput(SQF, [P,c]));
    V = W; T \= W);
  return (Vec(SQF));
}

/* Format of complex roots of polynomial P: [[z, c, rr], ...],
 * where z are the roots (up to conjugation if P is real),
 * c the multiplicity, and rr = 1 unless P is real and z
 * nonreal, in which case rr = 2. */
Polroots(P) =
{
  if (precision(P) == oo, return (Polrootsexact(P)));
  my(B = getlocalbitprec());
  /* P inexact. Error if we can't rule out multiple roots */
  if (exponent(polresultant(P, P')) < -B / poldegree(P),
    error("inexact and possibly nonsquarefree polynomial"));
  return ([[z, 1, 1] | z <- polroots(P)]);
}

/* P is an exact polynomial */
Polrootsexact(P) =
{ my(B = getlocalbitprec(), SQF = PolSQF(P), L = List());

  for (i = 1, #SQF,
    my([pol, c] = SQF[i], ro = polroots(pol));
    my(flre = (imag(pol) == 0));
    for (j = 1, #ro,
      my(z = ro[j], rr);
      rr = if (flre && exponent(imag(z)) > -B, 2, 1);
      listput(L, [z, c, rr]);
      /* conjugate roots adjacent in ro, skip conj(z) */
      if (rr == 2, j++)));
  return (Vec(L));
}

/* Partial fraction decomposition of rational function F with
 * denominator either exact or squarefree. The answer is a
 * vector of 4-component vectors [z, c, rr, sz], where z
 * ranges through the distinct poles of F, and only one among
 * two conjugates if the denominator DF of F is real, c is the
 * order of the pole, rr = 2 if DF is real and z nonreal, and
 * sz is the power series giving the polar decomposition
 * around z: if sz = sum(k = 0, v-1, ck * X^k) + O(x^c) then
 *   F = sum(k = 0, v-1, ck/(X-z)^(v-k)) + O(1).
 * The integer part is ignored. */
Ratdec(F) =
{ my(DF = denominator(F), NF = numerator(F), X = variable(F));
  my(L, Lconj, Poles);

  Poles = Polroots(DF);
  L = vector(#Poles); Lconj = List();
  for (i = 1, #Poles,
    my([z, c, rr] = Poles[i]);
    my(DFz = DF \ (X - z)^c, Qz = NF / DFz);
    my(sz = subst(Qz, X, X + z) + O(X^c));
    if (rr == 2 && imag(Qz),
      rr = 1; /* split into z and conj(z) */
      listput(Lconj, [conj(z), c, 1, conj(sz)]));
    L[i] = [z, c, rr, sz]);
  return (concat(L, Vec(Lconj)));
}
