/* Initialize weights and nodes for SumEulerMaclaurin2,
 * for functions evaluated to accuracy 2^(-B) in time O(B^om)
 * If 'a' is set use derivatives up to order k = a * B instead
 * of computing the optimal value */
/* Utility function: choose k, N */
SumEulerMaclaurin2kN(om = 2, a = 0) =
{ my(B = getlocalbitprec(), N, k);
  localbitprec(32);
  if (!a,
    my(C = (om+1)^(om+1) / om^om);
    a = log(2) / (1 + lambertw(2*Pi*C)));
  k = ceil(a * B); if (k % 2, k++);
  N = ceil(k / (2 * Pi * exp(1)) * 2^(B/k));
  return ([k, N]);
}
/* Return nodes, weights, N */
SumEulerMaclaurin2init(om = 2, a = 0) =
{ my([k, N] = SumEulerMaclaurin2kN(om, a), k2 = k \ 2);
  bernvec(k2); /* initialize Bernoulli cache */
  my([X, W, h] = DiffMinit(k, om));
  my(H, ih = 1 / h);
  H = powers(ih^2, k2 - 1, ih / (#X-1)!);
  X = [ N + h*v | v <- X ];
  W = sum(j = 1, k2, bernfrac(2*j)/(2*j) * H[j] * W[2*j]);
  return ([X, W, N]);
}

/* As SumEulerMaclaurin, computing the derivatives using
 * numerical differentiation */
SumEulerMaclaurin2(f, F = 0, XW = 0) =
{ my(B = getlocalbitprec(), S, X, W, N);

  [X, W, N] = if (XW, XW, SumEulerMaclaurin2init());
  F = Sumintinit(f, F);
  localbitprec(B + 32);
  S = sum(m = 1, N - 1, f(m), f(N) / 2.) - F(N);
  localbitprec(bitprecision(X[2]));
  S -= W * [f(x) | x <- X]~;
  return (bitprecision(S, B));
}
