/* Compute L(chi_D, s) by the Zetafast algorithm. */

Zetafastdata(q, T, B) =
{ my(qT = q * T, v = B \ 2, N, L, M0, M);

  localbitprec(32);
  N = max(round(5 * sqrt(qT) / B), 1);
  L = - log(factorial(v) >> B);  /* = log(2^B / v!) */
  M0 = ceil(qT * exp(L / v) / (2 * Pi * N));
  M = ceil(N * Solvedivlog(v, L));
  return ([v, N, M0, M]);
}

Vecsumchi(V, D) =
{
  if (D == 1, return (vecsum(V)));
  sum(m = 1, #V, kronecker(D, m) * V[m]);
}

ZetafastE1(D, s, data) =
{ my([v, N, M0] = data, q = abs(D), Z = I * q / (2 * Pi * N));
  my(sig = real(s), C, S, VE, E, VZ);

  VZ = powers(-Z, v - 1);
  VE = vector(M0, m, if (gcd(D, m) == 1, (m + Z)^s, 0));
  E = vector(M0, m, if (gcd(D, m) == 1, 1 / (m + Z), 0));
  S = Vecsumchi(dirpowers(M0, s), D);
  C = 1;
  for (w = 0, v - 1,
    S -= C * VZ[w + 1] * Vecsumchi(VE, D);
    if (w == v - 1, break);
    C *=  (s - w) / (w + 1);
    VE = vector(M0, m, VE[m] * E[m]));
  my(G = sqrt(q) * (2*Pi)^s * q^(-s - 1));
  if (D < 0, G *= -I);
  /* G = chi(-1) (2 pi)^(s-1) q^(-s) g(chi) */
  return (G * S * gamma(-s) * exp(-I * Pi/2 * s));
}

ZetafastQ(D, s, data) =
{ my([v, N, M0, M] = data, VS, VE, P, P0, S);

  VS = dirpowers(M, -s);
  VE = powers(exp(-1 / N), M);
  P = vector(v); P[v] = 1;
  forstep (w = v-1, 1, -1, P[w] = w * N * P[w+1]);
  P0 = P[1]; P = Polrev(P);
  S = if (D == 1,
    sum(n = 1, M, VS[n] * VE[n+1] * subst(P, 'x, n));
  ,
    sum(n = 1, M,
        my(k = kronecker(D, n));
        if (k, k * VS[n] * VE[n+1] * subst(P, 'x, n))));
  return (S / P0);
}

/* exp(z) avoiding underflow */
Exp(z) =
{ my(B = getlocalbitprec());
  localbitprec(32); my(E = (B + 20) * log(2));
  localbitprec(B); return (if (real(z) > -E, exp(z), 0.));
}

ZetafastG(s, data) =
{ my([v, N] = data);
  Exp(lngamma(s + v) + s * log(N) - log(s * (v - 1)!));
}

/* Valid when 1/2 <= Re(s) <= 1 */
Zetafast(s, D = 1) =
{ my(B = getlocalbitprec(), la, v, data, S);
  my(sig = real(s), T = imag(s), flconj = 0);

  if (sig < 1/2, error("please use the functional equation"));
  if (T == 0 && s == round(s) && s >= 0, return (lfun(D, s)));
  if (T < 0, flconj = 1; T = -T);
  [v] = data = Zetafastdata(abs(D), max(T,100), B);
  if (T < 3*v, return (lfun(D, s)));
  if (flconj, s = conj(s));
  localbitprec(B + 32);
  S = ZetafastQ(D, s, data) - if (D == 1, ZetafastG(1-s, data))
      + ZetafastE1(D, s-1, data);
  if (flconj, S = conj(S));
  return (bitprecision(S, B));
}
